/////////////////////////////////////////////////////////
//
// Gradient Normal Mapping Shader
//
// Copyright 2007 Shintaro Takemura
//

// O[oϐ //

// J[eNX`
texture g_baseTexture;

// @}bv
texture g_nmhTexture;

// J[Tv[
sampler2D tBase = sampler_state
{
	Texture = < g_baseTexture >;
	MinFilter = ANISOTROPIC;
	MagFilter = Linear;
	MipFilter = Linear;
	AddressU = Wrap;
	AddressV = Wrap;
	MaxAnisotropy = 16;
};

// @Tv[
sampler2D tNormalHeightMap = sampler_state
{
	Texture = < g_nmhTexture >;
	MinFilter = ANISOTROPIC;
	MagFilter = Linear;
	MipFilter = Linear;
	AddressU = Wrap;
	AddressV = Wrap;
	MaxAnisotropy = 16;
};

// x[XJ[O
float4 cBase;

// }bsO̗LtO(R=eNX`,G=@}bv,B=UVpj
float4 pMapping;

// ː(R=gU,G=͌,B=ˌ,A=˂̋)
float4 pReflect;

// Of[VJ[uiR=,G=j
float4 pGradient;

// Of[V
float4 vGradient;

// bṼXP[
float4 vScale;

// Cg̈ʒu
float4 vLightWS;

// Wϊs World
float4x4 mW;

// Wϊs World * View
float4x4 mWV;

// Wϊs World * View * Projection
float4x4 mWVP;

///////////////////////////////////////////////////////////////////////

// sNZVF[_[͍̓\
struct PS_INPUT
{
	float4 position          : POSITION;
	float3 texCoord          : TEXCOORD0;
	float3 vLightTS          : TEXCOORD1;
	float3 vViewTS           : TEXCOORD2;
	float3 vNormalWS         : TEXCOORD4;
	float3 vViewWS           : TEXCOORD5;
};

////////////////////////////////////////
// _VF[_[ //
PS_INPUT VS_TangentVectorSpace (
	float4 inPositionOS : POSITION,
	float2 inUV : TEXCOORD0,
	float3 vInTangentOS : TANGENT,
	float3 vInNormalOS : NORMAL,
	float3 vInBinormalOS : BINORMAL
	)
{
	// sNZVF[_[͍̓\̂錾
	PS_INPUT Out;

	// _W̃XP[O
	float4 vScaledPos = inPositionOS * vScale;

	// Wϊ̒_ʒuo
	Out.position = mul(vScaledPos, mWVP);

	if(pMapping.b<0.5)
	{
		// ڐxNgƈʒu̓ςtWo
		// ]@xNgƈʒu̓ςuWo
		// Of[VxNgƈʒu̓ςvWo
		Out.texCoord = float3(
			dot(vInTangentOS , inPositionOS) *  0.5 + 0.5,
			dot(vInBinormalOS, inPositionOS) * -0.5 + 0.5,
			dot(inPositionOS * sign(vScale), vGradient) );
	}
	else
	{
		// `ꂽtuŴ܂܎g
		Out.texCoord = float3(
			inUV.x,
			inUV.y,
			dot(inPositionOS * sign(vScale), vGradient) );
	}

	// O[oԂł̈ʒuZo
	float4 vPositionWS = mul( inPositionOS, mW );

	// O[oԂł̐ڋԊZo
	float3 vNormalWS   = mul( vInNormalOS  , (float3x3) mW );
	float3 vTangentWS  = mul( vInTangentOS , (float3x3) mW );
	float3 vBinormalWS = mul( vInBinormalOS, (float3x3) mW );

	// O[oԂł̎xNgo
	float3 vViewWS = -mul( inPositionOS, mWV );
	Out.vViewWS = vViewWS;

	// O[oԂł̖@xNgo
	Out.vNormalWS = vNormalWS;

	// ڐE@E]@𐳋K
	vNormalWS   = normalize( vNormalWS );
	vTangentWS  = normalize( vTangentWS );
	vBinormalWS = normalize( vBinormalWS );

	// ڋԍWϊsZo
	float3x3 mWorldToTangent = float3x3( vTangentWS, vBinormalWS, vNormalWS );

	// ڃxNgԂł̃CgxNgo
	Out.vLightTS = mul( vLightWS, mWorldToTangent );

	// ڃxNgԂł̎xNgo
	Out.vViewTS  = mul( mWorldToTangent, vViewWS  );

	return Out;
}

// sNZVF[_[ //
float4 PS_GradientNormalMapping( PS_INPUT i ) : COLOR0
{
	//  ԂꂽxNg𐳋K
	float3 vViewTS   = normalize( i.vViewTS  );
	float3 vLightTS  = normalize( i.vLightTS );

	// ڃxNgԂɂ@xNg`
	float3 vNormalTS = pMapping.y>=1 ? normalize( tex2D( tNormalHeightMap, i.texCoord ) * 2 - 1 ) : float3(0,0,1);
	
	// Of[VPxlZo
	float a = smoothstep(pGradient.x,pGradient.y,i.texCoord.z);
	float g = lerp(pGradient.z,pGradient.w,a);

	// J[eNX`eNZ擾
	float4 cTexelColor = pMapping.x>=1 ? tex2D( tBase, i.texCoord.xy ) : float4(1,1,1,1);

	// gUːZo
	float tDiffuse = saturate( dot( vNormalTS, vLightTS ));
	float d = (tDiffuse * pReflect.g + pReflect.r);

	// ʔːZo
	float s = 0;
	if(pReflect.b>0.00001)
	{
		float3 vReflectionTS = normalize( 2 * dot( vViewTS, vNormalTS ) * vNormalTS - vViewTS);
		float fRdotL = saturate( dot( vReflectionTS, vLightTS ));
		s = saturate( pow( fRdotL, pReflect.a )) * pReflect.b;
	}

	// ŏIo͌ʂZo
	return ( (float4(cBase.rgb * d,cBase.a) * cBase * cTexelColor + float4(s,s,s,0)) * float4(1,1,1,g) );
}

////////////////////////////////////////
// Of[V{@}bsO
technique GradientNormalMapping
{
	pass P0
	{
		VertexShader = compile vs_2_0 VS_TangentVectorSpace();
		PixelShader  = compile ps_2_0 PS_GradientNormalMapping();
	}
}

////////////////////////////////////////
// _VF[_[ //
void VS_General (
      float4 vPos    : POSITION,
      float2 vUV	 : TEXCOORD0,
	  float3 vNormal : NORMAL,
      out float4 oPos : POSITION,
      out float3 oTexCoord : TEXCOORD0,
	  out float4 oColor : COLOR0,
      out float4 oGloss : COLOR1
	)
{
	float4 vTPos = vPos * vScale;
	oPos = mul(vTPos, mWVP);	
	oTexCoord = float3(vUV.x,vUV.y,dot(vPos * sign(vScale), vGradient));

	float3 N = normalize( mul( vNormal, (float3x3)mWV ) );
    float LN = dot( N, float3(0,0,-1) );

	// gUːZo
	float d = (max(LN,0) * pReflect.g + pReflect.r);
	oColor.rgb = cBase.rgb * d; oColor.a = cBase.a;

	float s = (pReflect.b>0.0001) ? pReflect.b * pow(max(0,dot(normalize(float3(-1,1,-1)),N)),pReflect.a) : 0;
	oGloss = float4(s,s,s,0);
}

// sNZVF[_[ //
float4 PS_TextureMapping (
    float3 Tex0 : TEXCOORD0,
	float4 oColor : COLOR0,
	float4 oGloss : COLOR1
) : COLOR0
{
	// Of[VJ[Zo
	float a = smoothstep(pGradient.x, pGradient.y, Tex0.z);
	float g = lerp(pGradient.z,pGradient.w,a);

	// J[eNX`eNZ擾
	float4 cTexelColor = pMapping.x>=1 ? tex2D( tBase, Tex0.xy ) : float4(1,1,1,1);

	// ŏIo͌ʂZo
	return ( (oColor * cTexelColor + oGloss) * float4(1,1,1,g) );
}

////////////////////////////////////////
// eNX`}bsO
technique GeneralTextureMapping
{
	pass P0
	{
		VertexShader = compile vs_2_0 VS_General();
		PixelShader  = compile ps_2_0 PS_TextureMapping();
	}
}

////////////////////////////////////////
// ŒFhԂ

technique FixedColor
{
    pass P0
    {          
		VertexShader = null;
		PixelShader  = null;
    }
}

