//  Copyright (c) 2006, Gustavo Franco
//  Email:  gustavo_franco@hotmail.com
//  All rights reserved.

//  Redistribution and use in source and binary forms, with or without modification, 
//  are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer. 
//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution. 

//  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
//  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
//  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
//  PURPOSE. IT CAN BE DISTRIBUTED FREE OF CHARGE AS LONG AS THIS HEADER 
//  REMAINS UNCHANGED.

using System;
using System.IO;
using System.Data;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.ComponentModel;
using System.Drawing.Imaging;
using System.Collections.Generic;
using System.Runtime.InteropServices;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;


using CustomControls.Controls;

namespace CustomControls
{   
    public partial class Open3DFileDialog : OpenFileDialogEx
    {
        public enum Type
        {
            Texture,
            NormalMap,
            Mesh
        }

        // ۉ܂傤BBB
        IO.ImagePreviewRenderer renderer = null;
        IO.ModelPreviewRenderer rendererM = null;

        string textureFile;
        string title;
        Type type;

        #region Constructors
        public Open3DFileDialog(Type _type,string _textureFile)
        {
            InitializeComponent();
            textureFile = _textureFile;
            type = _type;
            title = this.Text;
        }
        #endregion

        #region Overrides
        /// <summary>
        /// t@CύXꂽƂ
        /// </summary>
        /// <param name="filePath">t@C</param>
        public override void OnFileNameChanged(string filePath)
        {
            if(!File.Exists(filePath)) return;
            FileAttributes  attr = File.GetAttributes(filePath);            
            if((attr & FileAttributes.Directory) == FileAttributes.Directory) return;

            switch(type)
            {
                case Type.Texture:
                    renderer.TextureImageFile = filePath;
                    break;
                case Type.NormalMap:
                    renderer.NormalMapFile = filePath;
                    break;
                case Type.Mesh:
                    rendererM.ModelFile = filePath;
                    break;
            }
        }

        /// <summary>
        /// tH_ύXꂽƂ
        /// </summary>
        /// <param name="folderName">tH_</param>
        public override void OnFolderNameChanged(string folderName)
        {
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            try
            {
                switch(type)
                {
                    case Type.Texture:
                    case Type.NormalMap:
                        renderer = new IO.ImagePreviewRenderer();
                        renderer.InitializeGraphics(pbxPreview);
                        if(type==Type.NormalMap)
                            renderer.TextureImageFile = textureFile;
                        break;
                    case Type.Mesh:
                        rendererM = new IO.ModelPreviewRenderer();
                        rendererM.InitializeGraphics(pbxPreview);
                        break;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("G[܂BfBXvC̐ݒĂB" + ex.ToString());
                return;
            }

            timer1.Start();
        }

        public override void OnClosingDialog()
        {
            timer1.Stop();
            switch(type)
            {
                case Type.Texture:
                case Type.NormalMap:
                    if(renderer!=null)
                    {
                        renderer.TerminateGraphics();
                        renderer = null;
                    }
                    break;
                case Type.Mesh:
                    if(rendererM!=null)
                    {
                        rendererM.TerminateGraphics();
                        rendererM = null;
                    }
                    break;
            }

            base.OnClosingDialog();
        }
        #endregion

        private void pbxPreview_Paint(object sender, PaintEventArgs e)
        {
            switch(type)
            {
                case Type.Texture:
                case Type.NormalMap:
                    if(renderer!=null)
                        renderer.Render(type==Type.NormalMap);
                    break;
                case Type.Mesh:
                    if(rendererM!=null)
                        rendererM.Render(type==Type.NormalMap);
                    break;
            }
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            switch(type)
            {
                case Type.Texture:
                case Type.NormalMap:
                    if(renderer!=null && renderer.isAvailable())
                        renderer.Render(type==Type.NormalMap);
                    break;
                case Type.Mesh:
                    if(rendererM!=null && rendererM.isAvailable())
                        rendererM.Render(type==Type.Mesh);
                    break;
            }
        }
    }
}