using System;
using System.IO;
using System.Collections.Generic;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace IO
{
    class DX9MeshGenerator
    {
        /// <summary>
        /// Direct3D 9pMesh𐶐
        /// </summary>
        /// <param name="device">Direct3D 9 foCX</param>
        /// <param name="m">bV[_[̃C^[tF[X</param>
        /// <param name="bComputeNormal">@vZ邩</param>
        /// <param name="bComputeTangent">ڃxNgԂvZ邩</param>
        /// <returns></returns>
        public Mesh generateMesh(Device device,IMeshLoader m, bool bComputeNormal, bool bComputeTangent)
        {
			Mesh mesh = null;
			DX9MultiVertex multivtx = new DX9MultiVertex(m.getIdealVertexFormat());

            // bV̐
			mesh = new Mesh(
				m.getTriangleCount(),
				m.getSeparatedVertexCount(),
				MeshFlags.Use32Bit,
				multivtx.getElms(),
				device );

            #region _obt@̐
            VertexBuffer vb = mesh.VertexBuffer;
            using (GraphicsStream vbuf = vb.Lock(0, 0, LockFlags.None))
            {
                VertexFlag vt = m.getIdealVertexFormat();
                for (int faceNo = 0; faceNo < m.getFaceCount(); faceNo++)
                {
                    int[] face = m.getFacePointArray(faceNo);
                    for (int apexNo = 0; apexNo < face.Length; apexNo++)
                    {
                        vbuf.Write(m.getPosition(face[apexNo]));

                        if ((vt & VertexFlag.WTMask) != 0)
                            vbuf.Write(m.getWeights(faceNo, apexNo));

                        if ((vt & VertexFlag.NORMAL) != 0)
                            vbuf.Write(m.getNormal(faceNo, apexNo));

                        if ((vt & VertexFlag.UVx1) != 0)
                            vbuf.Write(m.getTextureCoord(faceNo, apexNo));

                        if ((vt & VertexFlag.TANGENT) != 0)
                            vbuf.Write(m.getTangent(faceNo, apexNo));

                        if ((vt & VertexFlag.BINORMAL) != 0)
                            vbuf.Write(m.getBinormal(faceNo, apexNo));
                    }
                }
                vb.Unlock();
            }
            #endregion

            #region CfbNXobt@̐
            IndexBuffer ib = mesh.IndexBuffer;
            GraphicsStream ibuf = ib.Lock(0, 0, LockFlags.None);
            foreach (int[] indices in m.getSeparatedTriangles())
            {
                for (int a = 0; a < 3; a++)
                {
                    ibuf.Write(indices[a]);
                }
            }
            ib.Unlock();
            #endregion

            #region אږʂvZ
            // |Cg\f[^Zo
            int[] pointReps = new int[m.getSeparatedVertexCount()];
            for (int faceNo = 0, pointNo = 0; faceNo < m.getFaceCount(); faceNo++)
            {
                int[] face = m.getFacePointArray(faceNo);
                for (int apexNo = 0; apexNo < face.Length; apexNo++)
                {
                    pointReps[pointNo] = face[apexNo];
                    pointNo++;
                }
            }
            // |Cg\f[^烁bVאڏɕϊ
            int[] adjacencies = mesh.ConvertPointRepsToAdjacency(pointReps);

            // Op`̖@xNgZo
            List<Vector3> triNormals = new List<Vector3>();
            for (int f = 0; f < m.getFaceCount(); f++)
            {
                int[] vref = m.getFacePointArray(f);
                Vector3 vA = Vector3.Normalize(m.getPosition(vref[1]) - m.getPosition(vref[0]));
                Vector3 vB = Vector3.Normalize(m.getPosition(vref[2]) - m.getPosition(vref[0]));
                Vector3 normal = Vector3.Normalize(Vector3.Cross(vA,vB));
                for (int i = 2; i < vref.Length; i++)
                {
                    triNormals.Add(normal);
                }
            }

            // bVאڏɎ킦
            for(int a=0;a<adjacencies.Length;a++)
            {
                // ׂ荇ʂȂLZ
                if(adjacencies[a]==-1) continue;

                // ʂƖʂ̊pxZo
                int tA = a/3;
                int tB = adjacencies[a];
                float angle = (float)Math.Acos(Vector3.Dot(triNormals[tA],triNormals[tB]));

                // X[WOpxȏłΗאڂĂȂƂ݂Ȃ
                if(angle>m.getSmoothingAngle())
                {
                    adjacencies[a] = -1;
                }
            }
            #endregion

            #region @vZ
            if (bComputeNormal)
            {
                mesh.ComputeNormals(adjacencies);
            }
            #endregion

            #region ڃxNgԂvZ
            if (bComputeTangent)
            {
                mesh.ComputeTangent(0, 0, 0, 0, adjacencies);
            }
            #endregion

            #region obt@̐
            Int32[] attrBuf = mesh.LockAttributeBufferArray(LockFlags.None);
            int triCount = 0;
            for (int faceNo = 0; faceNo < m.getFaceCount(); faceNo++)
            {
                int[] face = m.getFacePointArray(faceNo);
                for (int apexNo = 2; apexNo < face.Length; apexNo++)
                {
                    attrBuf[triCount] = m.getMaterialIndex(faceNo);
                    triCount++;
                }
            }
            mesh.UnlockAttributeBuffer(attrBuf);
            #endregion

            // œK
            mesh.OptimizeInPlace(
                MeshFlags.OptimizeAttributeSort |
                MeshFlags.OptimizeVertexCache,
                adjacencies);

            return mesh;
        }
    }
}
