using System;
using System.IO;
using System.Collections.Generic;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace IO
{
    /// <summary>
    /// Direct3D 9 ėp_vfNX
    /// </summary>
    class DX9MultiVertex
    {
        public DX9MultiVertex(VertexFlag vt)
        {
            elms = new VertexElement[32];

            totalsize = 0;
            elmCount = 0;

            addElement(DeclarationType.Float3, DeclarationUsage.Position, 0);

            if ((vt & VertexFlag.WTMask) != 0)
            {
                int count = ((int)(vt & VertexFlag.WTMask) >> (int)VertexFlag.WTShift);

                for (int n = 0; n < count; n++)
                    addElement(DeclarationType.Float4, DeclarationUsage.BlendWeight, (byte)n);
            }

            if ((vt & VertexFlag.NORMAL) != 0)
                addElement(DeclarationType.Float3, DeclarationUsage.Normal, 0);


            if ((vt & VertexFlag.UVMask) != 0)
            {
                for (int n = 0; n < (int)(vt & VertexFlag.UVMask); n++)
                    addElement(DeclarationType.Float2, DeclarationUsage.TextureCoordinate, (byte)n);
            }

            if ((vt & VertexFlag.TANGENT) != 0)
                addElement(DeclarationType.Float3, DeclarationUsage.Tangent, 0);

            if ((vt & VertexFlag.BINORMAL) != 0)
                addElement(DeclarationType.Float3, DeclarationUsage.BiNormal, 0);

            elms[elmCount] = VertexElement.VertexDeclarationEnd;
            elmCount++;
        }

        public VertexElement[] getElms()
        {
            return elms;
        }

        public int getVertexSize()
        {
            return totalsize / 4;
        }

        private short totalsize = 0;
        private int elmCount = 0;
        private VertexElement[] elms = null;

        private void addElement(DeclarationType ctype, DeclarationUsage cusage, byte cindex)
        {
            elms[elmCount].Stream = 0;
            elms[elmCount].Offset = totalsize;
            elms[elmCount].DeclarationType = ctype;
            elms[elmCount].DeclarationMethod = DeclarationMethod.Default;
            elms[elmCount].DeclarationUsage = cusage;
            elms[elmCount].UsageIndex = cindex;
            switch (ctype)
            {
                case DeclarationType.Float1: totalsize += 4; break;
                case DeclarationType.Float2: totalsize += 8; break;
                case DeclarationType.Float3: totalsize += 12; break;
                case DeclarationType.Float4: totalsize += 16; break;
            }
            elmCount++;
        }
    }

}
