using System;
using System.IO;
using System.Collections.Generic;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;


// L\[XR[h 񂾂炟 MQOLoaderςĂ܂
// Copyright (C) 2001-2005 񂾂炟 
// http://www.wanderer.jp/dotnet/metasequoia.html

namespace Metasequoia
{
	/// <summary>
	/// G[O
	/// </summary>
	public class MQException : ApplicationException {
		public MQException(string msg) : base(msg) {
		}
	}

	public class MQLoadException : MQException {
		public MQLoadException() : base("MQ Load Exception") {
		}
	}

	public class MQInvalidFormatException : MQException {
		public MQInvalidFormatException() : base("MQ Invalid Format Exception") {
		}
	}

	public class MQNoSupportChunkException : MQException {
		public MQNoSupportChunkException() : base("MQ No Support Exception") {
		}
	}

	public class AsciiStringFileReader: IDisposable {
		private StreamReader sr;
		private List<string> scan;
		
		public AsciiStringFileReader(Stream stream) {
			sr = new StreamReader(stream, System.Text.Encoding.UTF8);
		}

		private void SetScan(ref string buf) {
			scan.Add(buf);
			buf = "";
		}

		private int Separate(string temp) {
			bool instr = false;
			string buf = "";

			int maxlen = temp.Length;
			for (int pos = 0; pos < maxlen; pos++) {
				char ch = temp[pos];
				switch (ch) {
				case ' ':
				case '\t':
					if (instr) goto ADDCHAR;
					if (buf.Length > 0) {
						SetScan(ref buf);
					}
					break;
				case '"':
					if (instr) {
						SetScan(ref buf);
						instr = false;
					} else {
						if (buf.Length > 0) {
							SetScan(ref buf);
						}
						instr = true;
					}
					break;
				default:
					if (instr) goto ADDCHAR;
					if ("(),:;=[]{}".IndexOf(ch) >= 0) {
						if (buf.Length > 0) {
							SetScan(ref buf);
						}
						buf = ch.ToString();
						SetScan(ref buf);
						continue;
					}
ADDCHAR:
					buf += ch;
					break;
				}

			}
			if (buf.Length > 0) {
				SetScan(ref buf);
			}

			return scan.Count;
		}

		public int GetLine() {
			string temp = sr.ReadLine();
            scan = new List<string>();
			if (temp == null) {
				return -1;
			}
			return Separate(temp);
		}

		public string GetString(int index) {
			return (string)scan[index];
		}

		public float ToFloat(int index) {
			return float.Parse((string)scan[index]);
		}

		public int ToInt(int index) {
			return int.Parse((string)scan[index]);
		}

		public uint ToUInt(int index) {
			return uint.Parse((string)scan[index]);
		}

		public void Dispose() {
			sr.Close();
		}

		public bool CheckCommand(string command, int optc) {
			return (optc <= scan.Count && string.Compare(command, (string)scan[0], true) == 0);
		}

		public bool CheckString(int index, string str) {
			return (index < scan.Count && string.Compare(str, (string)scan[index], true) == 0);
		}

		public void Skip(long offset) {
			long oldpos, newpos;

			oldpos = sr.BaseStream.Position;
			sr.BaseStream.Seek(offset, System.IO.SeekOrigin.Current);
			newpos = sr.BaseStream.Position;

			if (newpos - oldpos != offset) {
				throw new MQLoadException();
			}
		}
	}

/*
 	/// <summary>
	/// V[
	/// </summary>
	public struct MQScene {
		float posX, posY, posZ;
		float lookatX, lookatY, lookatZ;
		float head;
		float pitch;
		int ortho;
		float zoom2;
		MQColor ambient;
	}
*/

	/// <summary>
	/// F
	/// </summary>
	public struct MQColor {
		public float red, green, blue;
	}

	/// <summary>
	/// ގ
	/// </summary>
	public class MQMaterial {
		public string Name;				//	ގ
		public MQColor Color;			//	F
		public float Alpha;				//	x
		public float Diffuse;			//	gU
		public float Ambient;			//	͌
		public float Emission;			//	ȏƖ
		public float Specular;			//	ˌ
		public float Power;				//	ˌ̋
		public string TextureName;		//	͗l}bsO
		public string AlphaName;		//	}bsO
		public string BumpName;			//	ʃ}bsO

		public MQMaterial() {
            Color.red = 1;
            Color.green = 1;
            Color.blue = 1;
            Alpha = 1;
            Diffuse = 0.5f;
            Ambient = 0.5f;
            Emission = 0;
            Specular = 0;
            Power = 100;
            TextureName = "";
			AlphaName = "";
			BumpName = "";
		}

		public bool IsContainTexture() {
			if (TextureName != "") 
				return true;
			else
				return false;
		}
	}

	public class MQPoint {
		public float X, Y, Z;

		public MQPoint(float x, float y, float z) {
			X = x;
			Y = y;
			Z = z;
		}

		public void Set(float x, float y, float z) {
			X = x;
			Y = y;
			Z = z;
		}	
	}

	/// <summary>
	/// _
	/// </summary>
	public class MQVertext : MQPoint {
		private uint color;				//	_J[

		public MQVertext(float x, float y, float z) : base(x, y, z) {
			color = 0;
		}

		public MQPoint GetPoint() {
			return new MQPoint(X, Y, Z);
		}

		public Vector3 GetVector3() {
			return new Vector3(X, Y, Z);
		}

		public uint Color {
			get {
				return color;
			}
			set {
				color = value;
			}
		}

		//	_J[LȂtrue
		public bool IsColorActive {
			get {
				if ((color & 0xff000000) != 0)
					return true;
				else
					return false;
			}
		}
	}

	/// <summary>
	/// UVl
	/// </summary>
	public struct MQCoordinate {
		public float U, V;

		public MQCoordinate(float u, float v) {
			U = u;
			V = v;
		}
	}

	/// <summary>
	/// 
	/// </summary>
	public class MQFace {
		int vertexCount;				//	_
		int[] vertexIndex;					//	_CfbNX
		public int MaterialIndex;		//	}eACfbNX
		MQCoordinate[] UV;				//	UV

		public MQFace(int face) {
			init(face);
		}

		public MQFace(int face, int[] vi) {
			init(face);
			for(int v=0;v<face;v++)
				vertexIndex[v] = vi[v];

//			vi.CopyTo(vertexIndex,face);
//			vertexIndex = (int[])vi.Clone();
		}

		private void init(int vc) {
			if (vc != 2 && vc != 3 & vc != 4) {
				throw new MQException("_sł");
			}
			vertexCount = vc;
			vertexIndex = new int[vertexCount];
			UV = new MQCoordinate[vertexCount];
			MaterialIndex = -1;						//	}eAݒ
		}

		public void SetCoordinate(MQCoordinate[] uv) {
			for(int v=0;v<vertexCount;v++)
				UV[v] = uv[v];
		}

		public int VertexCount {
			get {
				return vertexCount;
			}
		}

		public int[] GetFacePointArray() {
			return vertexIndex;
		}

		public MQCoordinate[] GetUVArray() {
			return UV;
		}

		public MQFace Clone(int vertexOffset) {
			MQFace face = new MQFace(this.vertexCount, this.vertexIndex);
			for (int i = 0; i < face.vertexCount; i++) {
				face.vertexIndex[i] += vertexOffset;
			}
			face.MaterialIndex = this.MaterialIndex;
			face.UV = (MQCoordinate[])UV.Clone();
			int temp;
			MQCoordinate uvtemp;
			switch (face.vertexCount) {
			case 3:
				temp = face.vertexIndex[0];
				face.vertexIndex[0] = face.vertexIndex[2];
				face.vertexIndex[2] = temp;

				uvtemp = face.UV[0];
				face.UV[0] = face.UV[2];
				face.UV[2] = uvtemp;
				break;
			case 4:
				temp = face.vertexIndex[0];
				face.vertexIndex[0] = face.vertexIndex[3];
				face.vertexIndex[3] = temp;
				temp = face.vertexIndex[1];
				face.vertexIndex[1] = face.vertexIndex[2];
				face.vertexIndex[2] = temp;

				uvtemp = face.UV[0];
				face.UV[0] = face.UV[3];
				face.UV[3] = uvtemp;
				uvtemp = face.UV[1];
				face.UV[1] = face.UV[2];
				face.UV[2] = uvtemp;
				break;
			}
			return face;
		}
	}

	/// <summary>
	/// IuWFNg
	/// </summary>
	public class MQObject {
		public string Name;				//	IuWFNg
/*
 * 		int patch;						//	Ȗʂ̌`@0:ʁ@1:Ȗʃ^Cv1@2:Ȗʃ^Cv2@3:Catmull-Clark
 *		int segment;					//	Ȗʕ
 *		int shadeing;					//	VF[fBO@0:tbg@1:O[
 *		MQColor color;					//	F
 *		int colorType;					//	ӂ̐F^Cv@0:ݒF@1:IuWFNgŗLF
 */
        float facet;                    //  X[WOpx臒l
		int mirror;						//	ʃ^Cv@0:Ȃ@1:E𕪗@2:Eڑ
		int mirrorAxis;					//	ʓK@1:X@2:Y@4:Z
/*		float mirrorDistance;			//	ʐڑ
 *		int lathe;						//	]̃^Cv@0:Ȃ@3:
 *		int latheAxis;					//	]̂̎@0:X@1:Y@2:Z
 *		int latheSegment;				//	]̂̕
*/
        List<MQVertext> vertexList;			//	_(MQVertex List)
//		BVertex bvertex;				//	oCi_Q@iΉj
		List<MQFace> faceList;				//	


		/// <summary>
		/// RXgN^
		/// </summary>
		public MQObject() {
            vertexList = new List<MQVertext>();
            faceList = new List<MQFace>();
			mirror = 0;					//	ʂȂ
		}

		/// <summary>
		/// _Ԃ
		/// ۂɑ݂钸_ł͂ȂAz̑傫Ԃ
		/// </summary>
		/// <returns>_</returns>
		public  int GetVertexCount() {
			if (mirror != 0) {
				return vertexList.Count * 2;
			} else {
				return vertexList.Count;
			}
		}

		/// <summary>
		/// S_̕sړ
		/// </summary>
		/// <param name="tx">Xړ</param>
		/// <param name="ty">Yړ</param>
		/// <param name="tz">Zړ</param>
		public void Translate(float tx, float ty, float tz) {
			foreach (MQVertext v in vertexList) {
				v.X += tx;
				v.Y += ty;
				v.Z += tz;
			}
		}

		/// <summary>
		/// XP[ϊ
		/// </summary>
		/// <param name="scaleFactor">XP[OW</param>
		/// <param name="flipX">X]</param>
		/// <param name="flipY">Y]</param>
		/// <param name="flipZ">Z]</param>
		public void Scaling(float scaleFactor, bool flipX, bool flipY, bool flipZ) {
			foreach (MQVertext v in vertexList) {
				v.X *= scaleFactor;
				v.Y *= scaleFactor;
				v.Z *= scaleFactor;

				if (flipX) v.X *= -1;
				if (flipY) v.Y *= -1;
				if (flipZ) v.Z *= -1;
			}
		}

		/// <summary>
		/// X[WOpx臒l
		/// </summary>
		public float Facet {
			get {
				return facet;
			}
			set {
				facet = value;
			}
		}

		/// <summary>
		/// ʐݒ
		/// </summary>
		public int Mirror {
			get {
				return mirror;
			}
			set {
				mirror = value;
			}
		}

		/// <summary>
		/// ʓK
		/// </summary>
		public int MirroAxis {
			get {
				return mirrorAxis;
			}
			set {
				mirrorAxis = value;
			}
		}

		/// <summary>
		/// _̎擾
		/// </summary>
		/// <param name="index">擾_̃CfbNX</param>
		/// <returns>_f[^</returns>
		public MQVertext GetVertex(int index) {
			if (mirror == 0 || index < vertexList.Count) {
				return (MQVertext)vertexList[index];
			} else {	//	ʒ_
				MQVertext v = (MQVertext)vertexList[index - vertexList.Count];
				switch (mirrorAxis) {
				case 0:		//	X
					return new MQVertext(-v.X, v.Y, v.Z);
				case 1:		//	Y
					return new MQVertext(v.X, -v.Y, v.Z);
				default:	//	Z
					return new MQVertext(v.X, v.Y, -v.Z);
				}
			}
		}

		/// <summary>
		/// _ǉ
		/// </summary>
		/// <param name="v">ǉ钸_</param>
		public void AddVertex(MQVertext v) {
			vertexList.Add(v);
		}

		/// <summary>
		/// ʒǉ
		/// </summary>
		/// <param name="vc">_</param>
		/// <param name="vertexIndex">_ԍCfbNX</param>
		/// <returns>ʔԍ</returns>
		public int AddFace(int vc, int[] vertexIndex) {
			faceList.Add(new MQFace(vc, vertexIndex));
			return faceList.Count - 1;
		}

		/// <summary>
		/// FaceMaterialݒ肷
		/// </summary>
		/// <param name="cfi">_CfbNX</param>
		/// <param name="pal">}eACfbNX</param>
		/// <returns></returns>
		public bool SetFaceMaterial(int cfi, int pal) {
			((MQFace)faceList[cfi]).MaterialIndex = pal;
			return true;
		}

		/// <summary>
		/// FaceUVlݒ肷
		/// </summary>
		/// <param name="cfi">_CfbNX</param>
		/// <param name="uv">UVl</param>
		/// <returns></returns>
		public bool SetFaceCoordinateArray(int cfi, MQCoordinate[] uv) {
			((MQFace)faceList[cfi]).SetCoordinate(uv);
			return true;
		}

		/// <summary>
		/// ʐԂ
		/// </summary>
		/// <returns>ʐ</returns>
		public int GetFaceCount() {
			if (mirror == 0) 
			{
				return faceList.Count;
			} else {
				return faceList.Count * 2;
			}
		}

		/// <summary>
		/// _̎擾
		/// </summary>
		/// <param name="index">擾_̃CfbNX</param>
		/// <returns>_f[^</returns>
		public MQFace GetFace(int index) {
			if (mirror == 0 || index < faceList.Count) {
				return (MQFace)faceList[index];
			} else {
				return ((MQFace)faceList[index - faceList.Count]).Clone(vertexList.Count);
			}
		}

		/// <summary>
		/// w肵ʂɑ钸_Ԃ
		/// ɔjꂽ_w肳ꂽꍇ́A0Ԃ
		/// </summary>
		/// <param name="index">ʂ̃CfbNX</param>
		/// <returns>ʂ̒_</returns>
		public int GetFacePointCount(int index) {
			if (mirror != 0 && index >= faceList.Count) {
				index -= faceList.Count;
			}
			MQFace face = (MQFace)faceList[index];
			if (face != null) 
				return face.VertexCount;
			else
				return 0;
		}

		/// <summary>
		/// w肵ʂɊ蓖Ăꂽ}eÃCfbNXԂ
		/// FʂAʂjĂꍇ-1Ԃ
		/// </summary>
		/// <param name="index">ʂ̃CfbNX</param>
		/// <returns>}eÃCfbNX</returns>
		public int GetFaceMaterial(int index) {
			if (mirror != 0 && index >= faceList.Count) {
				index -= faceList.Count;
			}
			MQFace face = (MQFace)faceList[index];
			if (face != null) 
				return face.MaterialIndex;
			else
				return -1;
		}

		/// <summary>
		/// w肵_ɒ_J[ݒ肷
		/// </summary>
		/// <param name="index">_̃CfbNX</param>
		/// <param name="color">F</param>
		/// <returns></returns>
		public bool SetVertexColor(int index, uint color) {
			MQVertext vertex = this.GetVertex(index);
			vertex.Color = color;
			return true;
		}

		/// <summary>
		/// _J[̎擾
		/// </summary>
		/// <param name="index">_CfbNX</param>
		/// <returns></returns>
		public uint GetVertexColor(int index) {
			return ((MQVertext)vertexList[index]).Color;
		}

	}


	/// <summary>
	/// MQDocument ̊Tv̐łB
	/// </summary>
	public class MQDocument
	{
		private string m_name;					//	t@C
		private AsciiStringFileReader afs;
		private List<MQMaterial> MaterialList;			//	}eAXg
		private List<MQObject> ObjectList;			//	IuWFNgXg

		/// <summary>
		/// RXgN^
		/// </summary>
		public MQDocument() {
			m_name = "";
            MaterialList = new List<MQMaterial>();
            ObjectList = new List<MQObject>();
		}

		/// <summary>
		/// MQO@wb_`FbN
		/// </summary>
		/// <returns>wb_ȂtrueԂ</returns>
		private bool CheckHeader() {
			int sc = afs.GetLine();

			if (!afs.CheckCommand("Metasequoia", 2) || !afs.CheckString(1, "Document")) {
				throw new MQInvalidFormatException();
			}
			return true;
		}

		private bool LoadMaterial() {
			int i, scn;
			MQMaterial mat;

			while (true) {
				int sc = afs.GetLine();
				if (sc == -1) {
					throw new MQInvalidFormatException();
				}
				if (sc == 0) continue;

				if (afs.CheckCommand("}", 1)) break;

				mat = new MQMaterial();

				mat.Name = afs.GetString(0);

				for (scn = 1; scn < sc; ) {
					if (afs.CheckString(scn, "col") && scn + 7 <= sc) {
						if (!afs.CheckString(scn + 1, "(") || !afs.CheckString(scn + 6, ")")) break;
						mat.Color.red = afs.ToFloat(scn + 2);
						mat.Color.green = afs.ToFloat(scn + 3);
						mat.Color.blue = afs.ToFloat(scn + 4);
						mat.Alpha = afs.ToFloat(scn + 5);
						scn += 7;
					}
					else if (afs.CheckString(scn, "dif") && scn + 4 <= sc) {
						if (!afs.CheckString(scn + 1, "(") || !afs.CheckString(scn + 3, ")")) break;
						mat.Diffuse = afs.ToFloat(scn + 2);
						scn += 4;
					}
					else if(afs.CheckString(scn, "amb") && scn + 4 <= sc) {
						if(!afs.CheckString(scn + 1, "(") || !afs.CheckString(scn + 3, ")")) break;
						mat.Ambient = afs.ToFloat(scn+2);
						scn += 4;
					}
					else if(afs.CheckString(scn, "spc") && scn + 4 <= sc) {
						if(!afs.CheckString(scn + 1, "(") || !afs.CheckString(scn + 3, ")")) break;
						mat.Specular = afs.ToFloat(scn+2);
						scn += 4;
					}
					else if(afs.CheckString(scn, "emi") && scn + 4 <= sc) {
						if(!afs.CheckString(scn + 1, "(") || !afs.CheckString(scn + 3, ")")) break;
						mat.Emission = afs.ToFloat(scn + 2);
						scn += 4;
					}
					else if(afs.CheckString(scn, "power") && scn + 4 <= sc) {
						if(!afs.CheckString(scn + 1, "(") || !afs.CheckString(scn + 3, ")")) break;
						mat.Power = afs.ToFloat(scn + 2);
						scn += 4;
					}
					else if(afs.CheckString(scn, "tex") && scn + 4 <= sc) {
						if(!afs.CheckString(scn + 1, "(") || !afs.CheckString(scn + 3, ")")) break;
						mat.TextureName = afs.GetString(scn + 2);
						scn += 4;
					}
					else if(afs.CheckString(scn, "aplane") && scn + 4 <= sc) {
						if(!afs.CheckString(scn + 1, "(") || !afs.CheckString(scn + 3, ")")) break;
						mat.AlphaName = afs.GetString(scn + 2);
						scn += 4;
					}
					else if(afs.CheckString(scn, "bump") && scn + 4 <= sc) {
						if(!afs.CheckString(scn + 1, "(") || !afs.CheckString(scn + 3, ")")) break;
						mat.BumpName = afs.GetString(scn + 2);
						scn += 4;
					}
					else if(afs.CheckString(scn+1,"(")) {
						for(i = 2; i + scn < sc; i++) {
							if(afs.CheckString(scn + i, ")"))
								break;
						}
						scn += i + 1;
					} else
						break;
				}
				MaterialList.Add(mat);
			}
			return true;
		}

		/// <summary>
		/// Object`NVertex[h
		/// </summary>
		/// <param name="obj">ΏObject</param>
		/// <returns>[hɐtrue</returns>
		private bool LoadVertex(MQObject obj) {
			while (true) {
				int sc = afs.GetLine();
				if (sc == -1) {
					throw new MQInvalidFormatException();
				}

				if (afs.CheckCommand("}", 1)) break;

				if(sc < 3) continue;

				obj.AddVertex(new MQVertext(afs.ToFloat(0), afs.ToFloat(1), afs.ToFloat(2)));
			}
			return true;
		}

		/// <summary>
		/// Object`Nface[h
		/// </summary>
		/// <param name="obj">ΏObject</param>
		/// <returns></returns>
		private bool LoadFace(MQObject obj) {
			int scn;
			int vc,cfi;
			int[] vert_index = new int[4];

			while (true) {
				int sc = afs.GetLine();
				if (sc == -1) {
					throw new MQInvalidFormatException();
				}
				if (sc == 0) continue;

				if (afs.CheckCommand("}", 1)) break;

				vc = afs.ToInt(0);		//	_
				if(vc<=2) continue;
				scn = 1;
				if (afs.CheckString(scn,"V") && scn + vc + 3 <= sc) {
					if (!afs.CheckString(scn + 1, "(") || !afs.CheckString(scn + vc + 2, ")")) continue;
					if (vc >= 1) vert_index[0] = afs.ToInt(scn + 2);
					if (vc >= 2) vert_index[1] = afs.ToInt(scn + 3);
					if (vc >= 3) vert_index[2] = afs.ToInt(scn + 4);
					if (vc >= 4) vert_index[3] = afs.ToInt(scn + 5);
					cfi = obj.AddFace(vc, vert_index);
					scn += vc + 3;
				} else {
					throw new MQInvalidFormatException();
				}

				//	}eAw肪ꍇ
				if (afs.CheckString(scn,"M") && scn + 4 <= sc) {
					if(!afs.CheckString(scn + 1, "(") || !afs.CheckString(scn + 3, ")")) continue;
					int pal = afs.ToInt(scn + 2);
					if(0 <= pal && pal < this.GetMaterialCount())
						obj.SetFaceMaterial(cfi, pal);
					scn += 4;
				}

				//	UVw
				if(afs.CheckString(scn,"UV") && scn + vc * 2 + 3 <= sc) {
					MQCoordinate [] uv = new MQCoordinate[4];

					if(!afs.CheckString(scn+1,"(") || !afs.CheckString(scn + vc * 2 + 2, ")")) continue;
			
					if(vc >= 1) { uv[0].U = afs.ToFloat(scn + 2); uv[0].V = afs.ToFloat(scn + 3); }
					if(vc >= 2) { uv[1].U = afs.ToFloat(scn + 4); uv[1].V = afs.ToFloat(scn + 5); }
					if(vc >= 3) { uv[2].U = afs.ToFloat(scn + 6); uv[2].V = afs.ToFloat(scn + 7); }
					if(vc >= 4) { uv[3].U = afs.ToFloat(scn + 8); uv[3].V = afs.ToFloat(scn + 9); }

					obj.SetFaceCoordinateArray(cfi, uv);

					scn += vc * 2 + 3;
				}

				//	_J[
				if (afs.CheckString(scn, "COL") && scn + vc + 2 <= sc) {
					if(!afs.CheckString(scn + 1,"(") || !afs.CheckString(scn + vc + 2, ")")) continue;
			
					MQVertext vertex;
					if(vc >= 1) { vertex = obj.GetVertex(vert_index[0]); vertex.Color = afs.ToUInt(scn + 2); }
					if(vc >= 2) { vertex = obj.GetVertex(vert_index[1]); vertex.Color = afs.ToUInt(scn + 3); }
					if(vc >= 3) { vertex = obj.GetVertex(vert_index[2]); vertex.Color = afs.ToUInt(scn + 4); }
					if(vc >= 4) { vertex = obj.GetVertex(vert_index[3]); vertex.Color = afs.ToUInt(scn + 5); }
				}
			}
			return true;
		}

		private bool LoadObject(MQObject obj) {
			while (true) {
				int sc = afs.GetLine();
				if (sc == -1) {
					throw new MQInvalidFormatException();
				}

				if (afs.CheckCommand("}", 1)) break;

				else if(afs.CheckCommand("vertex", 3)) {
					if(!afs.CheckString(sc - 1,"{")) return false;
					/*if(sc >= 3)
						obj->ResizeVertexBuffer(afs->ToInt(1));*/
					if(!LoadVertex(obj)) return false;
				}
				else if(afs.CheckCommand("face", 2)) {
					if(!afs.CheckString(sc - 1, "{")) return false;
					/*if(sc >= 3)
						obj->ResizeFaceBuffer(afs->ToInt(1));*/
					if(!LoadFace(obj)) return false;
				}
				else if (afs.CheckCommand("mirror", 1)) {
					obj.Mirror = int.Parse(afs.GetString(1));
				}
				else if (afs.CheckCommand("facet", 1)) {
					obj.Facet = float.Parse(afs.GetString(1));
				}
				else {
					IgnoreUnknown(sc);
				}
			}
			return true;
		}

		private bool IgnoreBinary(int string_count) {
			long offset = afs.ToInt(string_count - 2);
			afs.Skip(offset);
			return true;
		}

		private bool IgnoreUnknown(int string_count) {
			if (string_count >= 1 && afs.CheckString(string_count - 1, "{")) {
				for (int comc=1; comc > 0 ; ) {
					string_count = afs.GetLine();
					if (string_count == -1) {
						throw new MQInvalidFormatException();
					}
					else if (string_count >= 1) {
						if (afs.CheckString(string_count - 1, "{")) {
							comc++;
						}
						else if (afs.CheckString(string_count - 1, "}")) {
							comc--;
						}
						else if (string_count>=3
							&& afs.CheckString(string_count - 3,"[")
							&& afs.CheckString(string_count - 1,"]")) {
							IgnoreBinary(string_count - 2);
						}
					}
				}
			}
			else if (string_count>=3 && afs.CheckString(string_count - 3, "[") && afs.CheckString(string_count - 1, "]")) {
				IgnoreBinary(string_count - 2);
			}

			return true;
		}

		/// <summary>
		/// MQO[hC[v
		/// </summary>
		/// <returns>Ƀ[hłtrueԂ</returns>
		private bool MainLoop() {
			while (true) {
				int sc = afs.GetLine();
				if (sc == -1) break;
				if (sc == 0) continue;

				if (afs.CheckCommand("TrialNoise", 2)) {
					throw new MQNoSupportChunkException();
				}
				else if (afs.CheckCommand("Material", 2)) {
					if (!afs.CheckString(sc - 1, "{")) {
						throw new MQInvalidFormatException();
					}
					if (!LoadMaterial()) {
						throw new MQInvalidFormatException();
					}
				}
				else if (afs.CheckCommand("Object",2)) {
					if(!afs.CheckString(sc - 1, "{")) {
						throw new MQInvalidFormatException();
					}

					MQObject curObj = new MQObject();
					
					if (sc >= 3) {
						curObj.Name = afs.GetString(1);
					}

					if (!LoadObject(curObj)) {
						throw new MQInvalidFormatException();
					}
					ObjectList.Add(curObj);
				}
				else {
					IgnoreUnknown(sc);
				}
			}
			return true;
		}

		/// <summary>
		/// Xg[[h
		/// </summary>
		/// <param name="fileName">t@C</param>
		public void Load(Stream stream) {
			using (afs = new AsciiStringFileReader(stream)) {
				if (!CheckHeader()) {
					throw new MQLoadException();
				}
				if (!MainLoop()) {
					throw new MQLoadException();
				}
			}
		}

		/// <summary>
		/// t@C[h
		/// </summary>
		/// <param name="fileName">t@C</param>
		public void Load(string fileName) {
            using(FileStream stream = new FileStream(fileName,FileMode.Open,FileAccess.Read,FileShare.Read))
            {
	            using (afs = new AsciiStringFileReader(stream)) {
		            if (!CheckHeader()) {
			            throw new MQLoadException();
		            }
		            if (!MainLoop()) {
			            throw new MQLoadException();
		            }
	            }
            }
		}

		/// <summary>
		/// t@CvpeB
		/// t@C{݂̂̂Ԃ
		/// </summary>
		public string Name {
			get {
				return Path.GetFileName(m_name);
			}
		}

		/// <summary>
		/// hLg̃IuWFNg̐Ԃ
		/// </summary>
		/// <returns>IuWFNg</returns>
		public int GetObjectCount() {
			return ObjectList.Count;
		}

		/// <summary>
		/// w肵CfbNX̃IuWFNg擾
		/// CfbNX͈͊OAIuWFNg폜ς݂̏ꍇ́AnullԂ
		/// </summary>
		/// <param name="index">擾IuWFNg̃CfbNX</param>
		/// <returns>擾IuWFNgւ̎Q</returns>
		public MQObject GetObject(int index) {
			if (index < 0 || index >= ObjectList.Count)
				return null;
			else
				return (MQObject)ObjectList[index];
		}

		/// <summary>
		/// hLg̃}eAԂ
		/// </summary>
		/// <returns>}eA</returns>
		public int GetMaterialCount() {
			return MaterialList.Count;
		}

		/// <summary>
		/// w肵CfbNX̃}eA擾
		/// CfbNX͈͊OA}eA폜ς݂̏ꍇ́AnullԂ
		/// </summary>
		/// <param name="index">擾}eÃCfbNX</param>
		/// <returns>擾}eAւ̎Q</returns>
		public MQMaterial GetMaterial(int index) {
			if (index < 0 || index >= MaterialList.Count)
				return null;
			else
				return (MQMaterial)MaterialList[index];
		}
	}
}
