/////////////////////////////////////////////////////////
//
// MQO Mesh Generator
//
// Copyright 2007 Shintaro Takemura
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using Metasequoia;

namespace IO
{    
    class MQOMeshLoader : IMeshLoader
    {
        /// <summary>
        /// _
        /// </summary>
        List<Vector3> points = new List<Vector3>();

        /// <summary>
        /// 
        /// </summary>
        List<int[]> faces = new List<int[]>();

        /// <summary>
        /// ʖ̂tuW
        /// </summary>
		List<Vector2[]>	faceUVs = null;

        /// <summary>
        /// ʖ̃}eAhc
        /// </summary>
        List<int> faceMaterialIndices = null;

        /// <summary>
        /// X[WOpx
        /// </summary>
        float smoothingAngle = 59.5f / 180.0f * (float)Math.PI;

        /// <summary>
        /// }eA
        /// </summary>
        public List<MQMaterial> materials = null;


        #region inteface for IMeshGenerator
        public int getVertexCount()
        {
            return points.Count;
        }

        public int getSeparatedVertexCount()
        {
            int t = 0;
            for (int f = 0; f < faces.Count; f++)
            {
                t += faces[f].Length;
            }
            return t;
        }

        public int getFaceCount()
        {
            return faces.Count;
        }

        public int[] getFacePointArray(int faceNo)
        {
            return faces[faceNo];
        }

        public Vector3 getPosition(int vertexNo)
        {
            return points[vertexNo];
        }

        public Vector2 getTextureCoord(int faceNo, int apexNo)
        {
            return faceUVs[faceNo][apexNo];
        }

        public float getSmoothingAngle()
        {
            return smoothingAngle;
        }

        public Vector3 getNormal(int faceNo, int apexNo)
        {
            int[] vref = (int[])faces[faceNo];
            return
                Vector3.Cross(
                    Vector3.Normalize(points[vref[1]] - points[vref[0]]),
                    Vector3.Normalize(points[vref[2]] - points[vref[0]]));
        }

        public Vector3 getTangent(int faceNo, int apexNo)
        {
            return Vector3.Empty;
        }

        public Vector3 getBinormal(int faceNo, int apexNo)
        {
            return Vector3.Empty;
        }

        public int getWeightLayersCount()
        {
            return 0;
        }

        public Vector4[] getWeights(int faceNo, int apexNo)
        {
            return null;
        }

        public int getMaterialIndex(int faceNo)
        {
            if (faceMaterialIndices == null) return 0;
            return faceMaterialIndices[faceNo];
        }

        /// <summary>
        /// Op`̗
        /// </summary>
        public IEnumerable<int[]> getSeparatedTriangles()
        {
            int vCount = 0;
            int[] idx = new int[3];
            for (int f = 0; f < faces.Count; f++)
            {
                int[] lvref = faces[f];
                if (lvref.Length < 3) continue;

                idx[0] = vCount + 0;
                idx[1] = vCount + 1;

                for (int i = 2; i < lvref.Length; i++)
                {
                    idx[2] = vCount + i;

                    yield return idx;
                    idx[1] = idx[2];
                }

                vCount += lvref.Length;
            }
        }

        /// <summary>
        /// Op`̗
        /// </summary>
        public IEnumerable<int[]> getTriangles()
        {
            int[] idx = new int[3];
            for (int f = 0; f < faces.Count; f++)
            {
                int[] vref = (int[])faces[f];

                if (vref.Length < 3) continue;

                idx[0] = vref[0];
                idx[1] = vref[1];

                for (int i = 2; i < vref.Length; i++)
                {
                    idx[2] = vref[i];

                    yield return idx;
                    idx[1] = idx[2];
                }
            }
        }

        /// <summary>
        /// Op`̐
        /// </summary>
        public int getTriangleCount()
        {
            int t = 0;
            for (int f = 0; f < faces.Count; f++)
            {
                t += faces[f].Length - 2;
            }
            return t;
        }


        /// <summary>
        /// KvƂȂ钸_tH[}bgԂ
        /// </summary>
        /// <returns></returns>
        public VertexFlag getIdealVertexFormat()
        {
            VertexFlag vertexformat = (VertexFlag)0;
            vertexformat |= VertexFlag.NORMAL;
            vertexformat |= VertexFlag.TANGENT;
            vertexformat |= VertexFlag.BINORMAL;
            vertexformat |= VertexFlag.UVx1;

            return vertexformat;
        }
        #endregion

        /// <summary>
        /// }eA̐Ԃ
        /// </summary>
        /// <returns></returns>
        public int getMaterialCount()
        {
            return materials.Count;
        }

        /// <summary>
        /// }eA擾
        /// </summary>
        /// <returns></returns>
        public MQMaterial getMaterial(int n)
        {
            if (materials.Count == 0)
                return new MQMaterial();
            return materials[n];
        }

        /// <summary>
        /// ό`\ȗ̂̃bV쐬
        /// </summary>
        public bool load(string modelFullPath)
        {
            try
            {
                using (FileStream stream = new FileStream(modelFullPath, FileMode.Open, FileAccess.Read, FileShare.Read))
                {
                    if (!load(stream)) return false;
                }
            }
            catch (System.Exception)
            {
                materials = null;
                return false;
            }
            return true;
        }

        public bool load(Stream stream)
        {
            Metasequoia.MQDocument mqo = new Metasequoia.MQDocument();
            mqo.Load(stream);

            points = new List<Vector3>();
            faces = new List<int[]>();
            faceUVs = new List<Vector2[]>();
            faceMaterialIndices = new List<int>();

            for (int oi = 0; oi < mqo.GetObjectCount(); oi++)
            {
                Metasequoia.MQObject obj = mqo.GetObject(oi);

                for (int fi = 0; fi < obj.GetFaceCount(); fi++)
                {
                    int vertexOffset = points.Count;

                    Metasequoia.MQFace face = obj.GetFace(fi);
                    int[] nface = new int[obj.GetFacePointCount(fi)];
                    for (int ai = 0; ai < nface.Length; ai++)
                        nface[ai] = face.GetFacePointArray()[ai] + vertexOffset;
                    faces.Add(nface);

                    Metasequoia.MQCoordinate[] fuvs = face.GetUVArray();
                    if (fuvs != null)
                    {
                        Vector2[] uvList = new Vector2[fuvs.Length];
                        for (int ai = 0; ai < fuvs.Length; ai++)
                        {
                            uvList[ai] = new Vector2(fuvs[ai].U, fuvs[ai].V);
                        }
                        faceUVs.Add(uvList);
                    }

                    smoothingAngle = obj.Facet / 180.0f * (float)Math.PI;

                    int materialIndex = face.MaterialIndex;
                    // }eAȂԂ݂͑Ă͂ȂȂ
                    if(materialIndex==-1) materialIndex = 0;
                    faceMaterialIndices.Add(materialIndex);
                }

                for (int vi = 0; vi < obj.GetVertexCount(); vi++)
                {
                    Metasequoia.MQVertext vertex = obj.GetVertex(vi);
                    points.Add(new Vector3(vertex.X, vertex.Y, -vertex.Z));
                }
                if (obj.GetVertexCount() == 0) return false;
            }

            materials = new List<Metasequoia.MQMaterial>();
            for (int mi = 0; mi < mqo.GetMaterialCount(); mi++)
            {
                materials.Add(mqo.GetMaterial(mi));
            }

            if (mqo.GetObjectCount() == 0) return false;
            return true;
        }

        public Vector3 normalizeScale()
        {
            Vector3 l = new Vector3(float.MaxValue, float.MaxValue, float.MaxValue);
            Vector3 u = new Vector3(float.MinValue, float.MinValue, float.MinValue);

            foreach (Vector3 v in points)
            {
                l = new Vector3(Math.Min(l.X, v.X), Math.Min(l.Y, v.Y), Math.Min(l.Z, v.Z));
                u = new Vector3(Math.Max(u.X, v.X), Math.Max(u.Y, v.Y), Math.Max(u.Z, v.Z));
            }
            Vector3 e = (u - l) * 0.5f;
            Vector3 c = (l + u) * 0.5f; ;
            float me = Math.Max(e.Z, Math.Max(e.X, e.Y));
            for (int n = 0; n < points.Count; n++)
            {
                if (me <= float.Epsilon) continue;
                points[n] = (points[n] - c) * (1.0f/me);
                points[n] = new Vector3(points[n].X, points[n].Y, points[n].Z/* - 1 + (e.Z) / me*/);
            }
            return e * (1.0f/me);
        }
    }
}
