/////////////////////////////////////////////////////////
//
// Model Preview Renderer
//
// Copyright 2007 Shintaro Takemura
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace IO
{
    public class ModelPreviewRenderer
    {
        /// <summary>
        /// Direct3D foCX
        /// </summary>
        private Device device = null;
        private PresentParameters presentParameters = null;

        // QctHg
        private Microsoft.DirectX.Direct3D.Font font = null;

        // J[eNX`
        private Dictionary<string,Texture> g_baseTextures = new Dictionary<string,Texture>();

        // @{nCg}bveNX`
        private Dictionary<string,Texture> g_nmhTextures = new Dictionary<string,Texture>();
        
        // |XgGtFNgpeNX`igpj
//        private Texture ds4Texture, ds16Texture, hBlurTexture, vBlurTexture;

        // r{[h̒_obt@i̋`j
        private CustomVertex.TransformedColored[] backgroundVertices = new CustomVertex.TransformedColored[4];

        // r{[h̃CfbNXobt@i̋`j
        private short[] backgroundIndices;

        // bV
        private Mesh mesh = null;

        // bV[_[
        private MQOMeshLoader loader = null;

        // GtFNg
        private Effect effect = null;

        // `̃XP[
        private Vector3 mScale;

        /// <summary>
        /// eNX`摜
        /// </summary>
        private string modelFile = null;

        public string ModelFile
        {
            get { return modelFile; }
            set {
                modelFile = value;
                if(device==null) return;
                generateMesh();
            }
        }

        /// <summary>
        /// bZ[W
        /// </summary>
        private string usage;

        public string Usage
        {
            get { return usage; }
            set { usage = value; }
        }

        /// <summary>
        /// p\
        /// </summary>
        /// <returns></returns>
        public bool isAvailable()
        {
            return device!=null && !device.Disposed;
        }

        /// <summary>
        /// OtBbNX֘Ã\[X
        /// </summary>
        public void InitializeGraphics(Control control)
        {
            presentParameters = new PresentParameters();

            presentParameters.Windowed = true;
            presentParameters.BackBufferWidth = control.ClientSize.Width;
            presentParameters.BackBufferHeight = control.ClientSize.Height;
            presentParameters.BackBufferFormat = Manager.Adapters.Default.CurrentDisplayMode.Format;
            presentParameters.BackBufferCount = 1;
            presentParameters.DeviceWindowHandle = control.Handle;
            presentParameters.SwapEffect = SwapEffect.Discard;

            presentParameters.EnableAutoDepthStencil = true;
            presentParameters.AutoDepthStencilFormat = DepthFormat.D24X8;

            // ł邾MSAAg
            for(MultiSampleType msType = MultiSampleType.EightSamples; msType>=MultiSampleType.None; msType--)
		    {
			    int result, quality;
			    if (Manager.CheckDeviceMultiSampleType(
                    Manager.Adapters.Default.Adapter,
				    DeviceType.Hardware,
                    Manager.Adapters[Manager.Adapters.Default.Adapter].CurrentDisplayMode.Format,
				    true,
                    msType,
                    out result, out quality))
			    {
				    presentParameters.MultiSample = msType;
				    presentParameters.MultiSampleQuality = 0;
                    break;
			    }
		    }

            presentParameters.FullScreenRefreshRateInHz = 0;
            presentParameters.PresentationInterval = PresentInterval.Default;

            // Direct3D foCX̐
            this.device = new Device(
                Manager.Adapters.Default.Adapter,
                DeviceType.Hardware,
                control.Handle,
                CreateFlags.HardwareVertexProcessing, presentParameters);

            // foCXZbg̃Cxgnh̐ݒ
            this.device.DeviceReset += this.OnResetDevice;

            // foCXXg̃Cxgnh̐ݒ
            this.device.DeviceLost += this.OnLostDevice;

            // foCXTCỸCxgnh̐ݒ
            this.device.DeviceResizing += OnDeviceResizing;

            string error;
            // GtFNgt@Cǂݍ
            ShaderFlags flag = ShaderFlags.None;
            effect = Effect.FromFile(
                this.device, Path.Combine(
                Path.GetDirectoryName(Application.ExecutablePath), @"3DDialog.fxo"),
                null, null, flag, null, out error);
            if (effect == null)
            {
                MessageBox.Show(error);
                return;
            }

            // \[Xmۂ̂߂ɈӐ}IɃfoCXZbgCxgs
            this.OnResetDevice(this,null);

        }

        /// <summary>
        /// OtBbNX֘A\[X
        /// </summary>
        public void TerminateGraphics()
        {
            if (this.device != null)
            {
                // \[XĴ߂ɈӐ}IɃfoCXZbgCxgs
                OnLostDevice(this, null);

                // foCXZbg̃Cxgnh̉
                this.device.DeviceReset -= this.OnResetDevice;

                // foCXXg̃Cxgnh̉
                this.device.DeviceLost -= this.OnLostDevice;

                // foCXTCỸCxgnh̉
                this.device.DeviceResizing -= OnDeviceResizing;

                // Direct3D foCX
                this.device.Dispose();
                this.device = null;

                // GtFNg
                effect.Dispose();
                effect = null;               
            }
        }

        /// <summary>
        /// foCXTCỸCxgnh
        /// </summary>
        private void OnDeviceResizing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            e.Cancel = true;
            return;
        }

        /// <summary>
        ///  foCXXg̃Cxgnh
        /// </summary>
        private void OnLostDevice(object sender, EventArgs e)
        {
			if(device!=null && effect!=null)
			{
                effect.OnLostDevice();
			}

            foreach(KeyValuePair<string,Texture> t in g_baseTextures)
            {
                t.Value.Dispose();
            }
            g_baseTextures.Clear();

            foreach (KeyValuePair<string, Texture> t in g_nmhTextures)
            {
                t.Value.Dispose();
            }
            g_nmhTextures.Clear();
/*
            if (ds4Texture != null)
                ds4Texture.Dispose();

            if (ds16Texture != null)
                ds16Texture.Dispose();

            if (hBlurTexture != null)
                hBlurTexture.Dispose();

            if (vBlurTexture != null)
                vBlurTexture.Dispose();
*/
            if (mesh != null)
            {
                mesh.Dispose();
                mesh = null;
            }

            if (font != null)
            {
                font.Dispose();
                font = null;
            }
        }

        /// <summary>
        ///  foCXZbg̃Cxgnh
        /// </summary>
        private void OnResetDevice(object sender, EventArgs e)
        {
            ///// tHg̍쐬 /////
            FontDescription fontDescription = new FontDescription();
            fontDescription.Height = 12;
            fontDescription.FaceName = "Arial";
            fontDescription.Quality = FontQuality.ClearType;

            // tHg쐬
            font = new Microsoft.DirectX.Direct3D.Font(this.device, fontDescription);

            ///// bV̍쐬 /////
            generateMesh();

            float width = device.PresentationParameters.BackBufferWidth;
            float height = device.PresentationParameters.BackBufferHeight;

            // r{[h̒_ECfbNXobt@쐬
            backgroundVertices[0].Position = new Vector4(0, 0, 1f, 1f);
            backgroundVertices[0].Color = Color.FromArgb(255, 64, 64, 64).ToArgb();

            backgroundVertices[1].Position = new Vector4(width, 0, 1f, 1f);
            backgroundVertices[1].Color = Color.FromArgb(255, 64, 64, 64).ToArgb();

            backgroundVertices[2].Position = new Vector4(0, height, 1f, 1f);
            backgroundVertices[2].Color = Color.FromArgb(255, 0, 0, 0).ToArgb();

            backgroundVertices[3].Position = new Vector4(width, height, 1f, 1f);
            backgroundVertices[3].Color = Color.FromArgb(255, 0, 0, 0).ToArgb();

            backgroundIndices = new short[] { 0, 1, 2, 1, 3, 2 };

            if(effect!=null)
            {
                effect.OnResetDevice();
            }
        }

        private bool generateMesh()
        {
            // MQOt@Cǂݍ
            MQOMeshLoader newLoader = new MQOMeshLoader();
            if (!newLoader.load(modelFile)) return false;
            mScale = newLoader.normalizeScale();

            Directory.SetCurrentDirectory(Path.GetDirectoryName(modelFile));

            DX9MeshGenerator gen = new DX9MeshGenerator();

            // \[Xj
            if (mesh != null)
                mesh.Dispose();

            foreach (KeyValuePair<string, Texture> t in g_baseTextures)
            {
                t.Value.Dispose();
            }
            g_baseTextures.Clear();

            foreach (KeyValuePair<string, Texture> t in g_nmhTextures)
            {
                t.Value.Dispose();
            }
            g_nmhTextures.Clear();

            mesh = gen.generateMesh(device, newLoader, true, true);
            for (int n = 0; n < newLoader.getMaterialCount(); n++)
            {
                // }eA擾
                Metasequoia.MQMaterial material = newLoader.getMaterial(n);

                if (!string.IsNullOrEmpty(material.TextureName) && File.Exists(material.TextureName))
                {
                    Texture tex = TextureLoader.FromFile(device, material.TextureName);
                    if(!g_baseTextures.ContainsKey(material.TextureName))
                        g_baseTextures.Add(material.TextureName, tex);
                }

                if (!string.IsNullOrEmpty(material.BumpName) && File.Exists(material.BumpName))
                {
                    Texture norm = TextureLoader.FromFile(device, material.BumpName);
                    if(!g_nmhTextures.ContainsKey(material.BumpName))
                        g_nmhTextures.Add(material.BumpName, norm);
                }
            }
            loader = newLoader;

            return true;
        }

        /// <summary>
        /// ɕ`R[h
        /// </summary>
        public void Render(bool bUseNormalMap)
        {
            if (this.device == null) return;

            // `eŃNAAyobt@NAB
            this.device.Clear(ClearFlags.ZBuffer | ClearFlags.Target,
                Color.Black, 1.0f, 0);


            // At@uhL
            this.device.RenderState.AlphaBlendEnable = true;
            this.device.RenderState.SourceBlend = Blend.SourceAlpha;
            this.device.RenderState.DestinationBlend = Blend.InvSourceAlpha;
            this.device.RenderState.ZBufferWriteEnable = true;
            this.device.RenderState.ZBufferEnable = true;
            this.device.RenderState.ZBufferFunction = Compare.LessEqual;
            this.device.RenderState.CullMode = Cull.CounterClockwise;
            this.device.RenderState.ShadeMode = ShadeMode.Gouraud;

            // GtFNgĂȂȂ炱̕`͒߂
            if(effect==null) return;

            // uBeginScenevƁuEndScenev̊Ԃɕ`eLq
            this.device.BeginScene();


            // eNjbNw
            effect.Technique = effect.GetTechnique("FixedColor");

            // r{[h̕\
            device.RenderState.ShadeMode = ShadeMode.Gouraud;
            device.RenderState.CullMode = Cull.None;
            device.VertexFormat = CustomVertex.TransformedColored.Format;
            device.DrawIndexedUserPrimitives(PrimitiveType.TriangleList,
                0, 6, 2, backgroundIndices, true, backgroundVertices);

    		long	m_psec,m_time;
            QueryPerformanceFrequency(out m_psec);
            QueryPerformanceCounter(out m_time);
            double passedSecond = (double)(m_time) / (double)m_psec;

            ///// [JWniIuWFNgj̐ݒ /////
            Matrix mW =
                Matrix.RotationY((float)passedSecond * 0.2f)
              * Matrix.RotationX(-(float)Math.PI * 0.15f);

            ///// O[oWni_j̐ݒ /////
            Matrix mV =
                Matrix.Translation(0,-1,32);
            Matrix mP =
                Matrix.PerspectiveFovLH(
                    Geometry.DegreeToRadian(20.0f),
                    (float)presentParameters.BackBufferWidth / (float)presentParameters.BackBufferHeight,
                    0.01f, 50.0f);

	        // O[oԂł̃Cgʒu
	        effect.SetValue("vLightWS", new Vector4(0,0,-1024,0) );
            
            // Of[V
            effect.SetValue("vGradient", new Vector4(0, 1, 0, 0));

            ////////////////////////////////////////////////// 
            // ˂bVɕ\

            // XP[ݒpϐ
            float vS = 3.0f;

            // XP[Zbgi傫߂Ɂj
            effect.SetValue("vScale", new Vector4( vS,-vS, vS, 1));

            // Of[VW
            effect.SetValue("pGradient", new Vector4(-0.5f*mScale.Y, mScale.Y, 0, 0.5f));

            // ː̐ݒ
            float fSpecular = bUseNormalMap ? 0.5f : 0.0f;
            effect.SetValue("pReflect", new Vector4(0.2f, 0.8f, fSpecular, 5.0f));

            // JO𔽎vɐݒ
            device.RenderState.CullMode = Cull.Clockwise;

            // ˂ꂽ{bNX̕`
            if (mesh != null)
                drawMesh(mesh, Matrix.Translation(0, -mScale.Y*vS, 0) * mW, mV, mP);

            ////////////////////////////////////////////////// 
            // bVɕ\

            // JO𔽎vɐݒ
            device.RenderState.CullMode = Cull.CounterClockwise;

            // Of[VW
            effect.SetValue("pGradient", new Vector4(1, 1, 1, 1));

            // XP[Zbgi傫߂Ɂj
            effect.SetValue("vScale", new Vector4(vS,vS,vS, 1));

            // {bNX̕`
            if (mesh != null)
                drawMesh(mesh, Matrix.Translation(0, mScale.Y*vS, 0) * mW, mV, mP);

            // ̕\
            if(!string.IsNullOrEmpty(usage))
            {
                font.DrawText(null, usage, 8, presentParameters.BackBufferHeight-20, Color.White);
            }

            this.device.EndScene();


            #region V[̍XV
            try
            {
                this.device.Present();
            }
            // foCX̃`FbNƃZbg
            catch (DeviceLostException)
            {
                int result;
                // Zbg\ǂ`FbN
                if (!this.device.CheckCooperativeLevel(out result))
                {
                    // Zbg\Ȃ烊Zbg
                    if (result == (int)ResultCode.DeviceNotReset)
                        this.device.Reset(presentParameters);
                    // ܂ZbgłȂ΁A΂炭X[v
                    else if (result == (int)ResultCode.DeviceLost)
                        System.Threading.Thread.Sleep(20);
                }
            }
            #endregion
        }

        /// <summary>
        /// bV̕`
        /// </summary>
        /// <param name="mesh">bV</param>
        /// <param name="wM">[hWn</param>
        /// <param name="vM">r[Wn</param>
        /// <param name="pM">eWn</param>
        /// <param name="scale">XP[</param>
        protected void drawMesh(Mesh mesh,Matrix wM,Matrix vM,Matrix pM)
        {
            // Wϊs̐ݒ
            effect.SetValue("mW", wM);
            effect.SetValue("mWV", wM * vM);
            effect.SetValue("mWVP", wM * vM * pM);

            // _Oi`j
            for (int n = 0; n < loader.getMaterialCount(); n++)
            {
                // }eA擾
                Metasequoia.MQMaterial mt = loader.getMaterial(n);

                // }eA񂩂eNX`擾
                Texture tex, nmap;
                bool bHasTexture = g_baseTextures.TryGetValue(mt.TextureName, out tex);
                bool bHasNormalMap = g_nmhTextures.TryGetValue(mt.BumpName, out nmap);

                // eNjbNw
                if(bHasNormalMap)
                    effect.Technique = effect.GetTechnique("GradientNormalMapping");
                else
                    effect.Technique = effect.GetTechnique("GeneralTextureMapping");

                // }bsOtO
                effect.SetValue("pMapping", new Vector4(bHasTexture ? 1 : 0, bHasNormalMap ? 1 : 0, 1, 0));

                // eNX`VF[_[ɃZbg
                effect.SetValue("g_baseTexture", bHasTexture ? tex : null);

                // @}bvVF[_[ɃZbg
                effect.SetValue("g_nmhTexture", bHasNormalMap ? nmap : null);

                // F̐ݒ
                effect.SetValue("cBase", new Vector4(mt.Color.red, mt.Color.green, mt.Color.blue, mt.Alpha));

                // ː̐ݒ
                effect.SetValue("pReflect", new Vector4(
                    mt.Ambient * 0.25f + mt.Emission,
                    mt.Diffuse,
                    mt.Specular,
                    mt.Power));


                int nPasses = 1;
                nPasses = effect.Begin(0);

                for (int i = 0; i < nPasses; i++)
                {
                    effect.BeginPass(i);
                    effect.CommitChanges();
                    mesh.DrawSubset(n);
                    effect.EndPass();                    
                }

                effect.End();

            }
        }

        #region Win32 API
        [System.Security.SuppressUnmanagedCodeSecurity]
        [DllImport("kernel32")]
        public static extern bool QueryPerformanceFrequency(out long PerformanceFrequency);

        [System.Security.SuppressUnmanagedCodeSecurity]
        [DllImport("kernel32")]
        public static extern bool QueryPerformanceCounter(out long PerformanceCount);
        #endregion
    }
}
